# coding=utf-8
import numpy as np
from tirt import BaseModel


def irt_consistency_score(model, theta=None):
    # irt评分一致性，检测乱答，随机作答的被试
    if not isinstance(model, BaseModel):
        raise ValueError('mode must be bayes_probit or bayes_logistic or ml_probit or ml_logistic')

    if theta is not None:
        prob_val = model.prob(theta)
    else:
        theta = model.solve
        prob_val = model.prob(theta)

    score = model.score

    c_ratio = 1.0 * (len(prob_val[score == 1][prob_val[score == 1] > 0.5]) +
                     len(prob_val[score == 0][prob_val[score == 0] < 0.5])) / len(prob_val)
    return c_ratio


def ctt_consistency_score(score):
    # ctt评分一致性，检测乱答，随机作答的被试

    if not isinstance(score, np.ndarray):
        raise ValueError('mode must be bayes_probit or bayes_logistic or ml_probit or ml_logistic')

    return np.std(score)



# coding=utf-8
import random
import numpy as np
from settings import TRIPLETS_PERMUTATION


def _gen_item_pattern(trait_size, items_size_per_dim, block_size=3):
    """
    生成迫选问卷试题排列组合，呈现形式为维度的排列
    :param trait_size: int, 特质数量
    :param items_size_per_dim: int, 每个特质的题量
    :return: list(int), 模拟正式试题的维度排列
    """
    item_pool = []
    for i in range(items_size_per_dim):
        item_pool.extend(range(trait_size))
    items = []
    while True:
        if not item_pool:
            break
        _dim_list = np.random.choice(item_pool, block_size)
        if len(set(_dim_list)) == block_size:
            # 每个题块的试题均要属于不同维度
            for dim in _dim_list:
                item_pool.remove(dim)
            items.extend(list(_dim_list))
    return items


def _gen_item_dt(items):
    """
    生成试题字典，key为试题顺序，value为试题所在维度
    :param items: list(int), 试题维度排列列表
    :return: dict(int: int) 试题字典
    """
    item_dt = {}
    for i, item in enumerate(items):
        item_dt[i] = item
    return item_dt


def gen_item_dict(trait_size, items_size_per_dim, block_size=3):
    items = _gen_item_pattern(trait_size, items_size_per_dim, block_size)
    item_dt = _gen_item_dt(items)
    print u'试题生成成功！'
    return item_dt


def gen_item_bank(trait_size, item_size, block_size=3, lower=1, upper=4, avg=0, std=1):
    """
    生成用于自适应测验的题库
    :param trait_size: int
    :param item_size: int
    :param block_size: int
    :param lower: int|float
    :param upper: int|float
    :param avg: int|float
    :param std: int|float
    :return: 
    """
    if not isinstance(trait_size, int):
        raise ValueError('trait_size must be int')
    if not isinstance(item_size, int):
        raise ValueError('trait_size must be int')
    if block_size not in (2, 3):
        raise ValueError('block_size must be 2 or 3')
    if not isinstance(lower, (int, float)):
        raise ValueError('lower must be int or float')
    if not isinstance(upper, (int, float)):
        raise ValueError('upper must be int or float')
    if not isinstance(avg, (int, float)):
        raise ValueError('avg must be int or float')
    if not isinstance(std, (int, float)):
        raise ValueError('std must be int or float')

    trait_list = range(trait_size)
    item_bank = []
    for i in range(item_size):
        _item_list = np.random.choice(trait_list, block_size, False)
        _item_dt = _gen_item_dt(_item_list)
        params = random_params(_item_dt, trait_size, block_size=block_size, lower=lower, upper=upper, avg=avg, std=std)
        item_bank.append({'dim': _item_list, 'params': params})
    return np.array(item_bank)


def _sim_score(block_nums):
    """
    生成随机试题得分（配对比较得分模式）
    :param block_nums: 题库的数量
    :return: 配对比较得分列表
    """
    score = []
    for j in range(block_nums):
        _score = random.choice(TRIPLETS_PERMUTATION)
        score.extend(_score)
    score = np.array(score)
    return score


def _get_block_nums(trait_size, items_size_per_dim, block_size=3):
    return trait_size * items_size_per_dim / block_size


def sim_scores(trait_size, items_size_per_dim, score_nums):
    """
    生成多个随机试题得分（配对比较得分模式）
    :param items_size_per_dim: int
    :param trait_size: int
    :param score_nums: int, 分数数量
    :return: list(ndarray), 多个得分列表
    """
    if not isinstance(trait_size, int):
        raise ValueError('trait_size must be int')
    if not isinstance(items_size_per_dim, int):
        raise ValueError('items_size_per_dim must be int')
    if not isinstance(score_nums, int):
        raise ValueError('score_nums must be int')

    block_nums = _get_block_nums(trait_size, items_size_per_dim)
    return [_sim_score(block_nums) for _ in range(score_nums)]


def _get_triplet_ipsative_score(pair_score, trait_size, item_dt, item_sign):
    """
    配对比较分数转自比分数
    :param pair_score: list(int), 配对比较分数
    :param trait_size: int, 特质的数量
    :param item_dt: dict(int: int), 试题维度字典
    :param item_sign: list(1|-1), 试题正反向列表
    :return: list(int), 自比分数列表
    """
    score = [0 for i in range(trait_size)]
    for j in range(len(pair_score) / 3):
        i1i2 = pair_score[j * 3]
        i1i3 = pair_score[j * 3 + 1]
        i2i3 = pair_score[j * 3 + 2]
        i1 = 1
        i2 = 0
        i3 = -1
        if i1i2 == 0:
            i1, i2 = i2, i1
        if i1i3 == 0:
            i1, i3 = i3, i1
        if i2i3 == 0:
            i2, i3 = i3, i2
        score[item_dt[j * 3]] += i1 * item_sign[item_dt[j * 3]]
        score[item_dt[j * 3 + 1]] += i2 * item_sign[item_dt[j * 3 + 1]]
        score[item_dt[j * 3 + 2]] += i3 * item_sign[item_dt[j * 3 + 2]]
    return score


def _pair_random_params(item_dt, trait_size, lower=1, upper=4, avg=0, std=1):
    """
    生成block_size为2的多维随机斜率（区分度）,和一维阈值（通俗度）
    :param std: 多维正态分布的标准差
    :param avg: 多维正态分布的期望值
    :param item_dt: dict(int:int), 试题字典
    :param trait_size: int，特质数量
    :param lower: int(>0), uniform分布的下界
    :param upper: int(>lower), uniform分布的上界
    :return: tuple(ndarray, ndarray), 斜率和阈值
    """
    keys = item_dt.keys()
    pair_nums = len(keys) / 2
    keys.sort()
    a = np.zeros((pair_nums, trait_size))
    a1 = np.random.uniform(lower, upper, pair_nums * 2)
    a2 = np.random.uniform(lower, upper, pair_nums * 2)

    for i in range(pair_nums):
        i1 = item_dt[2 * i]
        i2 = item_dt[2 * i + 1]
        a[i][i1] = a1[i]
        a[i][i2] = a2[2 * i] * -1
    b = np.random.normal(avg, std, pair_nums)
    return a, b


def _triplet_random_params(item_dt, trait_size, lower=1, upper=4, avg=0, std=1):
    """
    生成block_size为3的多维随机斜率（区分度）,和一维阈值（通俗度）
    :param std: 多维正态分布的标准差
    :param avg: 多维正态分布的期望值
    :param item_dt: dict(int:int), 试题字典
    :param trait_size: int，特质数量
    :param lower: int(>0), uniform分布的下界
    :param upper: int(>lower), uniform分布的上界
    :return: tuple(ndarray, ndarray), 斜率和阈值
    """
    keys = item_dt.keys()
    pair_nums = len(keys)
    keys.sort()
    a = np.zeros((pair_nums, trait_size))
    a1 = np.random.uniform(lower, upper, pair_nums)
    a2 = np.random.uniform(lower, upper, pair_nums)

    for i in range(len(keys) / 3):
        i1 = item_dt[3 * i]
        i2 = item_dt[3 * i + 1]
        i3 = item_dt[3 * i + 2]
        a[3 * i][i1] = a1[3 * i]
        a[3 * i][i2] = a2[3 * i] * -1
        a[3 * i + 1][i1] = a1[3 * i + 1]
        a[3 * i + 1][i3] = a2[3 * i + 1] * -1
        a[3 * i + 2][i2] = a1[3 * i + 2]
        a[3 * i + 2][i3] = a2[3 * i + 2] * -1
    b = np.random.normal(avg, std, pair_nums)
    return a, b


def random_params(item_dt, trait_size, block_size=3, lower=1, upper=4, avg=0, std=1):
    """
    生成随机参数
    :param item_dt: dict,试题字典，例如题块为3的0:1,1:0,2:2}代表第1题的第一个陈述测的是特质1，
    第二个陈述测的是特质0，第三个陈述测的是特质2
    :param trait_size: int
    :param block_size: int
    :param lower: int|float
    :param upper: int|float
    :param avg: int|float
    :param std: int|float
    :return: 
    """
    if not isinstance(item_dt, dict):
        raise ValueError('item_dt must be dict')
    if not isinstance(trait_size, int):
        raise ValueError('trait_size must be int')
    if block_size not in (2, 3):
        raise ValueError('block_size must be 2 or 3')
    if not isinstance(lower, (int, float)):
        raise ValueError('lower must be int or float')
    if not isinstance(upper, (int, float)):
        raise ValueError('upper must be int or float')
    if not isinstance(avg, (int, float)):
        raise ValueError('avg must be int or float')
    if not isinstance(std, (int, float)):
        raise ValueError('std must be int or float')

    if block_size == 3:
        return _triplet_random_params(item_dt, trait_size, lower=lower, upper=upper, avg=avg, std=std)
    elif block_size == 2:
        return _pair_random_params(item_dt, trait_size, lower=lower, upper=upper, avg=avg, std=std)


class cached_property(object):
    """
    # 从django抄的详见同名函数
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, type=None):
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


# coding=utf-8
import numpy as np
from scipy.stats import norm
from exception import ConvergenceError, ItemParamError, ScoreError, ThetaError, IterMethodError
from utils import cached_property


class BaseModel(object):

    # 最大牛顿迭代次数
    _newton_max_iter = 1000
    # 牛顿迭代步长
    _newton_step_size = 0.1
    # 最大梯度上升次数
    _gradient_max_iter = 1000
    # 梯度上升步长
    _gradient_step_size = 0.01
    # 参数估计精度
    _tol = 1e-5

    def __init__(self, slop, threshold, init_theta=None, score=None, iter_method='newton'):
        """
        不管是probit还是logit，都是用一样的参数估计算法，
        基于牛顿迭代的极大似然算法和贝叶斯最大后验算法
        :param slop: ndarray(float), 多维向量，斜率，区分度
        :param threshold: ndarray(float), 单维向量，阈值，通俗度，难度
        :param init_theta: ndarray(int|float), 特质向量初值
        :param score:  ndarray(0|1), 得分向量
        """
        if not isinstance(slop, np.ndarray):
            raise ItemParamError('item param must be ndarray')
        if not isinstance(threshold, np.ndarray):
            raise ItemParamError('item param must be ndarray')

        if len(slop.shape) == 1:
            slop.shape = 1, slop.shape[0]

        if len(slop) != len(threshold):
            raise ItemParamError('item param must be same length')

        if score is not None:
            if not isinstance(score, np.ndarray):
                raise ScoreError('score must be ndarray')
            if len(score) != len(slop):
                raise ScoreError('score must be same length as item param')

        if init_theta is not None and not isinstance(init_theta, np.ndarray):
            raise ThetaError('init_theta must be ndarray')

        if iter_method not in ('newton', 'gradient_ascent'):
            raise IterMethodError('iter_method must be newton or gradient_ascent')

        self._slop = slop
        self._score = score
        self._threshold = threshold
        self._init_theta = init_theta if init_theta is not None else np.zeros(len(self._slop[0]))
        # 默认bayes先验正态分布标准差
        # TODO 改为根据样本估计
        self._inv_psi = np.identity(len(self._slop[0]))
        self._iter_method = iter_method

    @property
    def score(self):
        return self._score

    def _prob(self, theta):
        raise NotImplementedError

    def prob(self, theta):
        # 回答为1的概率
        if not isinstance(theta, np.ndarray):
            raise ThetaError('theta must be ndarray')
        return self._prob(theta)

    def _z(self, theta):
        """
        probit和logit的z值
        :param theta: ndarray(int|float), 特质向量初值
        :return: ndarray(float), z值向量
        """
        return np.sum(self._slop * theta, 1) - self._threshold

    def z(self, theta):
        if not isinstance(theta, np.ndarray):
            raise ThetaError('theta must be ndarray')
        return self._z(theta)

    def _get_hessian_n_jacobian(self, theta):
        """
        抽象方法，目的是返回海塞矩阵和雅克比一阶导数向量，用于牛顿迭代
        :param theta: ndarray(int|float), 特质向量初值
        """
        raise NotImplementedError

    def _get_jacobian(self, theta):
        """
        抽象方法，目的是返回雅克比向量（矩阵），用于梯度上升
        :param theta: 
        """
        raise NotImplementedError

    @property
    def newton(self):
        """
        基于牛顿迭代的参数估计
        :return: ndarray(int|float), 特质向量初值
        """
        theta0 = self._init_theta * 1.0
        for i in range(self._newton_max_iter):
            hes, jac = self._get_hessian_n_jacobian(theta0)
            temp = self._newton_step_size * np.dot(hes, jac)
            theta = theta0 - temp
            if np.max(np.abs(temp)) < self._tol:
                # print i
                return np.round(theta, 3)
            theta0 = theta
        raise ConvergenceError('no convergence')

    @property
    def gradient_ascent(self):
        # 梯度上升参数估计
        theta0 = self._init_theta * 1.0
        for i in range(self._gradient_max_iter):
            jac = self._get_jacobian(theta0)
            theta = theta0 + self._gradient_step_size * jac
            if np.max(np.abs(self._gradient_step_size * jac)) < self._tol:
                return np.round(theta, 3)
            theta0 = theta
        raise ConvergenceError('no convergence')

    @cached_property
    def solve(self):
        return getattr(self, self._iter_method)


class BaseLogitModel(BaseModel):

    # D值
    D = 1.702

    def _prob(self, theta):
        """
        答1的概率值
        :param theta: ndarray(int|float), 特质向量初值
        :return:  ndarray(float)，作答为1概率的向量值
        """
        e = np.exp(self.D * self._z(theta))
        return e / (1.0 + e)

    def _dloglik(self, theta, prob_val):
        """
        logistic对数似然函数的一阶导数
        :param theta: ndarray(int|float), 特质向量初值
        :param prob_val: ndarray(float)，作答为1概率的向量值
        :return: 
        """
        return self.D * np.dot(self._slop.transpose(), self._score -prob_val)

    def _expect(self, prob_val):
        return self.D ** 2 * np.dot(self._slop.transpose() * prob_val * (1 - prob_val), self._slop)

    def _ddloglik(self, theta, prob_val):
        """
        logistic对数似然函数的二阶导数
        :param theta: ndarray(int|float), 特质向量初值
        :param prob_val: ndarray(float)，作答为1概率的向量值
        :return:
        """
        return -1 * self._expect(prob_val)

    def info(self, theta):
        # 信息矩阵
        if not isinstance(theta, np.ndarray):
            raise ThetaError('theta must be ndarray')
        prob_val = self._prob(theta)
        return self._expect(prob_val)


class MLLogitModel(BaseLogitModel):

    # 极大似然估计

    def _get_hessian_n_jacobian(self, theta):
        prob_val = self._prob(theta)
        hes = np.linalg.inv(self._ddloglik(theta, prob_val))
        jac = self._dloglik(theta, prob_val)
        return hes, jac

    def _get_jacobian(self, theta):
        prob_val = self._prob(theta)
        return self._dloglik(theta, prob_val)


class BayesLogitModel(BaseLogitModel):

    # 贝叶斯modal估计

    def _bayes_dloglik(self, theta, prob_val):
        # 贝叶斯modal的一阶导数
        return self._dloglik(theta, prob_val) - theta

    def _bayes_ddloglik(self, theta, prob_val):
        # 贝叶斯modal的二阶导数
        return self._ddloglik(theta, prob_val) - self._inv_psi

    def _get_hessian_n_jacobian(self, theta):
        prob_val = self._prob(theta)
        hes = np.linalg.inv(self._bayes_ddloglik(theta, prob_val))
        jac = self._bayes_dloglik(theta, prob_val)
        return hes, jac

    def _get_jacobian(self, theta):
        prob_val = self._prob(theta)
        return self._bayes_dloglik(theta, prob_val)

    def info(self, theta):
        # 信息矩阵
        if not isinstance(theta, np.ndarray):
            raise ThetaError('theta must be ndarray')
        _info = super(BayesLogitModel, self).info(theta)
        return _info + self._inv_psi


class BaseProbitModel(BaseModel):

    # probit基础模型

    def _prob(self, theta):
        # probit概率值
        return norm.cdf(self._z(theta))

    def _h(self, z):
        # probit函数的h值，方便计算
        return (1.0 / ((2 * np.pi) ** 0.5)) * np.exp(-1 * z ** 2 / 2.0)

    def _w(self, h, prob):
        # probit函数的w值，可以看成权重，方便计算和呈现

        pq = (1 - prob) * prob
        return h ** 2 / (pq + 1e-10)

    def _get_h_prob_val_w(self, theta):
        z = self._z(theta)
        h = self._h(z) + 1e-10
        prob_val = self._prob(theta)
        w = self._w(h, prob_val)
        return h, prob_val, w

    def _dloglik(self, theta, prob_val, h, w):
        # probit一阶导数
        return np.dot(self._slop.transpose(), w * (self._score - prob_val) / h)

    def _ddloglik(self, theta, w):
        # probit二阶导数
        return -1 * np.dot(self._slop.transpose() * w, self._slop)

    def info(self, theta):
        # 信息矩阵
        if not isinstance(theta, np.ndarray):
            raise ThetaError('theta must be ndarray')
        h, prob_val, w = self._get_h_prob_val_w(theta)
        return np.dot(self._slop.transpose() * w, self._slop)


class MLProbitModel(BaseProbitModel):

    # probit极大似然估计

    def _get_hessian_n_jacobian(self, theta):
        h, prob_val, w = self._get_h_prob_val_w(theta)
        hes = np.linalg.inv(self._ddloglik(theta, w))
        jac = self._dloglik(theta, prob_val, h, w)
        return hes, jac

    def _get_jacobian(self, theta):
        h, prob_val, w = self._get_h_prob_val_w(theta)
        return self._dloglik(theta, prob_val, h, w)


class BayesProbitModel(BaseProbitModel):

    # 贝叶斯modal估计

    def _bayes_dloglik(self, theta, prob_val, h, w):
        return self._dloglik(theta, prob_val, h, w) - theta

    def _bayes_ddloglik(self, theta, w):
        return self._ddloglik(theta, w) - self._inv_psi

    def _get_hessian_n_jacobian(self, theta):
        h, prob_val, w = self._get_h_prob_val_w(theta)
        hes = np.linalg.inv(self._bayes_ddloglik(theta, w))
        jac = self._bayes_dloglik(theta, prob_val, h, w)
        return hes, jac

    def _get_jacobian(self, theta):
        h, prob_val, w = self._get_h_prob_val_w(theta)
        return self._bayes_dloglik(theta, prob_val, h, w)

    def info(self, theta):
        # 信息矩阵
        if not isinstance(theta, np.ndarray):
            raise ThetaError('theta must be ndarray')
        _info = super(BayesProbitModel, self).info(theta)
        return _info + self._inv_psi


def estimate_theta(slop, threshold, score, init_theta=None):
    # TODO 参数估计API
    pass



from sim import SimFixedTirt, SimAdaptiveTirt
from tirt import BayesLogitModel, BayesProbitModel, MLLogitModel, MLProbitModel
from utils import gen_item_dict, gen_item_bank, sim_scores, random_params
from cs import irt_consistency_score, ctt_consistency_score



# coding=utf-8
import math
import numpy as np
from exception import UnknownModelError, IterMethodError
from utils import gen_item_dict, random_params, cached_property, gen_item_bank
from tirt import BayesProbitModel, MLProbitModel, MLLogitModel, BayesLogitModel
from cs import irt_consistency_score


class BaseSimTirt(object):

    MODEL = {'bayes_probit': BayesProbitModel, 'ml_probit': MLProbitModel,
             'bayes_logit': BayesLogitModel, 'ml_logit': MLLogitModel}

    def __init__(self, subject_nums, trait_size, model='bayes_probit',
                 iter_method='newton', block_size=3, lower=1, upper=4, avg=0, std=1):
        """

        :param subject_nums: int, 模拟被试的人数
        :param trait_size: int, 特质数量
        :param iter_method: str
        :param model: str, 模型
        :param block_size: int, 题块 
        :param lower: int|float
        :param upper: int|float
        :param avg: int|float
        :param std: int|float
        """
        if not isinstance(subject_nums, int):
            raise ValueError('subject_nums must be int')
        if not isinstance(trait_size, int):
            raise ValueError('trait_size must be int')
        if model not in ('bayes_probit', 'bayes_logistic', 'ml_probit', 'ml_logistic'):
            raise ValueError('mode must be bayes_probit or bayes_logistic or ml_probit or ml_logistic')
        if block_size not in (2, 3):
            raise ValueError('block_size must be 2 or 3')
        if not isinstance(lower, (int, float)):
            raise ValueError('lower must be int or float')
        if not isinstance(upper, (int, float)):
            raise ValueError('upper must be int or float')
        if not isinstance(avg, (int, float)):
            raise ValueError('avg must be int or float')
        if not isinstance(std, (int, float)):
            raise ValueError('std must be int or float')
        if iter_method not in ('newton', 'gradient_ascent'):
            raise IterMethodError('iter_method must be newton or gradient_ascent')

        self._subject_nums = subject_nums
        self._trait_size = trait_size
        self._block_size = block_size
        self._lower = lower
        self._upper = upper
        self._avg = avg
        self.std = std
        self._iter_method = iter_method
        self._model = self._get_model(model)

    def _get_model(self, model):
        try:
            return self.MODEL[model]
        except KeyError:
            raise UnknownModelError('unknown model, must be "bayes_probit" or '
                                    '"ml_probit" or "bayes_logit" or "ml_logit"')

    @cached_property
    def random_thetas(self):
        """
        生成特质向量
        :return: ndarray
        """
        return np.random.multivariate_normal(np.zeros(self._trait_size),
                                             np.identity(self._trait_size), self._subject_nums)

    def _get_init_theta(self):
        return np.zeros(self._trait_size)

    def _get_mean_error(self, theta_list):
        return np.mean(np.abs(theta_list - self.random_thetas))


class SimFixedTirt(BaseSimTirt):

    # 模拟固定迫选测验

    def __init__(self, items_size_per_dim, *args, **kwargs):
        super(SimFixedTirt, self).__init__(*args, **kwargs)
        # 每个维度的题量
        self._items_size_per_dim = items_size_per_dim
        self.theta_list = []

    @cached_property
    def _item_dict(self):
        return gen_item_dict(self._trait_size, self._items_size_per_dim, self._block_size)

    @cached_property
    def random_params(self):
        """
        生成随机参数
        :return: tuple(ndarray, ndarray)
        """
        return random_params(self._item_dict, self._trait_size,
                             block_size=self._block_size, lower=self._lower,
                             upper=self._upper, avg=self._avg,
                             std=self.std)

    @cached_property
    def scores(self):
        """
        生成作答记录
        :return: list(ndarray)
        """
        slop, threshold = self.random_params
        thetas = self.random_thetas
        score_list = []
        for theta in thetas:
            p_list = self._model(slop, threshold).prob(theta)
            score = np.random.binomial(1, p_list, len(p_list))
            score_list.append(score)
        return score_list

    def sim(self):
        # 模拟作答
        slop, threshold = self.random_params
        init_theta = self._get_init_theta()
        theta_list = []
        for score in self.scores:
            # print score
            model = self._model(slop, threshold, init_theta, score, self._iter_method)
            try:
                res = model.solve
                theta_list.append(res)
            except Exception as e:
                theta_list.append(np.nan)
        mean_error = self._get_mean_error(np.array(theta_list))
        print '模拟结束，平均误差{0}'.format(mean_error)
        self.theta_list = theta_list
        return theta_list

    def get_consistency_scores(self):
        slop, threshold = self.random_params
        c_score_list = []
        for i, theta in enumerate(self.theta_list):
            model = self._model(slop, threshold, score=self.scores[i])
            c_score_list.append(irt_consistency_score(model, theta))
        return c_score_list


class SimAdaptiveTirt(BaseSimTirt):

    def __init__(self, item_size, max_sec_item_size=10, *args, **kwargs):
        super(SimAdaptiveTirt, self).__init__(*args, **kwargs)
        # 题库题量
        self._item_size = item_size
        # 已做答试题编号保存记录
        self._has_answered_item_idx = {}
        # 已做答得分保存记录
        self._score = {}
        # 参数估计保存记录
        self._theta = {}
        # 作答试题斜率保存记录
        self._slop = {}
        # 作答试题阈值保存记录
        self._threshold = {}
        # 第二阶段最大答题次数
        self._max_sec_item_size = max_sec_item_size

    @property
    def scores(self):
        return self._score

    @property
    def thetas(self):
        return self._theta

    def _add_slop(self, theta_idx, slop):
        if theta_idx in self._slop:
            self._slop[theta_idx] = np.concatenate((self._slop[theta_idx], slop))
        else:
            self._slop[theta_idx] = slop

    def _get_slop(self, theta_idx):
        return self._slop[theta_idx]

    def _get_threshold(self, theta_idx):
        return self._threshold[theta_idx]

    def _add_threshold(self, theta_idx, threshold):
        if theta_idx in self._threshold:
            self._threshold[theta_idx] = np.concatenate((self._threshold[theta_idx], threshold))
        else:
            self._threshold[theta_idx] = threshold

    def _add_answered_item_idx(self, theta_idx, used_item_idx_list):
        if theta_idx in self._has_answered_item_idx:
            self._has_answered_item_idx[theta_idx].extend(used_item_idx_list)
        else:
            self._has_answered_item_idx[theta_idx] = used_item_idx_list

    def _get_answered_item_idx_set(self, theta_idx):
        return set(self._has_answered_item_idx[theta_idx])

    def _get_can_use_items(self, theta_idx):
        can_use_idx = self._get_can_use_idx(theta_idx)
        return self.item_bank[list(can_use_idx)]

    def _get_can_use_idx(self, theta_idx):
        can_use_idx = self._item_idx_set - self._get_answered_item_idx_set(theta_idx)
        return can_use_idx

    def _add_score(self, theta_idx, score):
        if theta_idx in self._score:
            self._score[theta_idx] = np.concatenate((self._score[theta_idx], score))
        else:
            self._score[theta_idx] = score

    def _get_score(self, theta_idx):
        return self._score[theta_idx]

    def _add_theta(self, theta_idx, theta):
        if theta_idx in self._theta:
            self._theta[theta_idx].append(theta)
        else:
            self._theta[theta_idx] = [theta]

    def _get_theta(self, theta_idx):
        return self._theta[theta_idx][-1]

    @cached_property
    def item_bank(self):
        return gen_item_bank(self._trait_size, self._item_size, self._block_size)

    @cached_property
    def _item_idx_set(self):
        return set(range(self._item_size))

    def _get_random_choice_items(self, theta_idx):
        rand_choice_size = self._get_random_choice_size()
        while True:
            items = []
            dims = []
            used_idx_list = []
            idx_list = np.random.choice(list(self._item_idx_set), rand_choice_size, False)
            for i in idx_list:
                item = self.item_bank[i]
                items.append(item)
                dims.extend(item['dim'])
                used_idx_list.append(i)
            # if len(set(dims)) == self._trait_size:
            self._add_answered_item_idx(theta_idx, used_idx_list)
            return items

    def _get_random_choice_size(self):
        return int(math.ceil(1.0 * self._trait_size / self._block_size))

    def _get_random_choice_params(self, theta_idx):
        first_rand_items = self._get_random_choice_items(theta_idx)
        slop = []
        threshold = []
        for item in first_rand_items:
            slop.extend(item['params'][0])
            threshold.extend(item['params'][1])
        return np.array(slop), np.array(threshold)

    def _first_random(self, theta, theta_idx):
        # 第一阶段，随机抽题
        slop, threshold = self._get_random_choice_params(theta_idx)
        p_list = self._model(slop, threshold).prob(theta)
        score = np.random.binomial(1, p_list, len(p_list))
        init_theta = self._get_init_theta()
        model = self._model(slop, threshold, init_theta, score, self._iter_method)
        theta = model.solve
        self._add_score(theta_idx, score)
        self._add_theta(theta_idx, theta)
        self._add_slop(theta_idx, slop)
        self._add_threshold(theta_idx, threshold)

    def _second_random(self, theta, theta_idx):
        item = self._get_next_item(theta_idx)

        score = self._get_next_score(item, theta, theta_idx)
        # print score

        est_theta = self._get_estimate_theta(score, theta_idx)
        # print est_theta
        # print np.mean(np.abs(est_theta - theta))
        self._add_theta(theta_idx, est_theta)
        return est_theta

    def _get_estimate_theta(self, score, theta_idx):
        # 参数估计
        now_slop = self._get_slop(theta_idx)
        now_threshold = self._get_threshold(theta_idx)
        init_theta = self._get_init_theta()
        model = self._model(now_slop, now_threshold, init_theta, score, self._iter_method)
        est_theta = model.solve
        return est_theta

    def _get_next_score(self, item, theta, theta_idx):
        # 模拟自适应抽题的下一题得分
        item_slop = item['params'][0]
        self._add_slop(theta_idx, item_slop)
        item_threshold = item['params'][1]
        self._add_threshold(theta_idx, item_threshold)
        p_list = self._model(item_slop, item_threshold).prob(theta)
        item_score = np.random.binomial(1, p_list, len(p_list))
        self._add_score(theta_idx, item_score)
        score = self._get_score(theta_idx)
        return score

    def _get_next_item(self, theta_idx):
        # 获得自适应抽题的下一道题
        est_theta = self._get_theta(theta_idx)
        items = self._get_can_use_items(theta_idx)
        slop = self._get_slop(theta_idx)
        threshold = self._get_threshold(theta_idx)
        test_info = self._model(slop, threshold).info(est_theta)
        info_list = []
        for _item in items:
            _slop, _threshold = _item['params']
            item_info = self._model(_slop, _threshold).info(est_theta)
            info_list.append(np.linalg.det(test_info + item_info))
        max_info_idx = np.array(info_list).argmax()
        item = items[max_info_idx]
        idx = list(self._get_can_use_idx(theta_idx))[max_info_idx]
        self._add_answered_item_idx(theta_idx, [idx])
        return item

    def sim(self):
        thetas = self.random_thetas
        theta_list = []
        for i, theta in enumerate(thetas):
            try:
                est_theta = np.nan
                self._first_random(theta, i)
                for j in range(self._max_sec_item_size):
                    est_theta = self._second_random(theta, i)
                print u'第{0}个被试模拟成功！'.format(i + 1)
            except Exception as e:
                print e
                continue
            theta_list.append(est_theta)
        mean_error = self._get_mean_error(np.array(theta_list))
        print '模拟结束，平均误差{0}'.format(mean_error)
        return theta_list







# coding=utf-8

"""
tirt
====

the simulation of Thurstone Item Response Theory, include fixed forced
test and adaptive forced test.
模拟瑟斯顿项目反应理论，包括固定测验和自适应测验。

瑟斯顿IRT模型简介和应用
-----------------------

瑟斯顿IRT模型主要应用于迫选式非认知测验（人格测验，动机测验，兴趣测验等）。

固定测验模拟
------------

模拟100个被试，30个维度，每个维度10个陈述，每道题3个陈述，所以下面这个陈述总共有100题

::

    from tirt import SimFixedTirt

    fixed_tirt = SimFixedTirt(subject_nums=100, trait_size=30, items_size_per_dim=10)
    theta_list = fixed_tirt.sim()
    score_list = fixed_tirt.scores

    for i, theta in enumerate(theta_list):
        print score_list[i]
        print theta

自适应测验模拟
--------------

模拟1个被试，题库600道题，30个维度，首先随机抽10题，第二阶段抽合适的题40道题，总共50道题

::

    from tirt import SimAdaptiveTirt

    sat = SimAdaptiveTirt(subject_nums=1, item_size=600, trait_size=30, max_sec_item_size=40)
    sat.sim()

    for key, value in sat.thetas.items():
        print sat.scores[key]
        print value

一致性
------

迫选测验通常都没有测谎量表（迫选测验本身抗作假），而衡量被试是否认真作答有更好的一致性分数

::

    from tirt import irt_consistency_score, sim_scores, BayesProbitModel, gen_item_dict, SimFixedTirt
    from tirt.utils import random_params

    # 生成试题字典
    item_dict = gen_item_dict(30, 10, block_size=3)
    # 生成试题参数
    a, b = random_params(item_dict, 30, block_size=3)
    # 生成随机得分
    scores = sim_scores(30, 10, 10)

    for score in scores:
        model = BayesProbitModel(a, b, score=score)
        # 打印一致性
        print irt_consistency_score(model)

    model = SimFixedTirt(trait_size=30, items_size_per_dim=10, subject_nums=100, model='bayes_probit')
    model.sim()
    print model.get_consistency_scores()
"""


from setuptools import setup

setup(
    name='tirt',
    version='0.0.4',
    packages=['tirt'],
    url='https://github.com/inuyasha2012/tirt',
    license='MIT',
    author='inuyasha2012',
    author_email='inuyasha021@163.com',
    description='the simulation of Thurstone Item Response Theory, include fixed forced test and adaptive forced test. ',
    long_description=__doc__,
    install_requires=['numpy', 'scipy'],
    classifiers=[
        'Programming Language :: Python :: 2.7',
    ]
)



class ConvergenceError(Exception):
    """
    no convergence
    """


class UnknownModelError(Exception):
    """
    unknown model
    """


class ItemParamError(Exception):
    """
    Item Param Type Error
    """


class ScoreError(Exception):
    """
    score error
    """


class ThetaError(Exception):
    """
    theta error
    """


class IterMethodError(Exception):
    """
    iter method error
    """